import json
import sys
import uuid
from functools import partial

import eventlet
from cpl_core.dependency_injection import ServiceProviderABC
from eventlet import wsgi
from flask import Flask, request, jsonify, Response, make_response
from flask_cors import CORS
from flask_socketio import SocketIO

from bot_api.configuration.api_settings import ApiSettings
from bot_api.configuration.frontend_settings import FrontendSettings
from bot_api.exception.service_exception import ServiceException
from bot_api.logging.api_logger import ApiLogger
from bot_api.model.error_dto import ErrorDTO
from bot_api.route.route import Route


class Api(Flask):

    def __init__(
            self,
            logger: ApiLogger,
            services: ServiceProviderABC,
            api_settings: ApiSettings,
            frontend_settings: FrontendSettings,
            *args, **kwargs
    ):
        if not args:
            kwargs.setdefault('import_name', __name__)

        Flask.__init__(self, *args, **kwargs)

        self._logger = logger
        self._services = services
        self._apt_settings = api_settings

        self._cors = CORS(self, support_credentials=True)

        # register before request
        self.before_request_funcs.setdefault(None, []).append(self.before_request)
        exc_class, code = self._get_exc_class_and_code(Exception)
        self.error_handler_spec[None][code][exc_class] = self.handle_exception

        # websockets
        self._socketio = SocketIO(self, cors_allowed_origins='*')
        self._socketio.on_event('connect', self.on_connect)
        self._socketio.on_event('disconnect', self.on_disconnect)

    def _register_routes(self):
        for path, f in Route.registered_routes.items():
            route = f[0]
            kwargs = f[1]
            cls = None
            qual_name_split = route.__qualname__.split('.')
            if len(qual_name_split) > 0:
                cls_type = vars(sys.modules[route.__module__])[qual_name_split[0]]
                cls = self._services.get_service(cls_type)

            partial_f = partial(route, self if cls is None else cls)
            partial_f.__name__ = route.__name__
            self.route(path, **kwargs)(partial_f)

    def handle_exception(self, e: Exception):
        self._logger.error(__name__, f'Caught error', e)

        if isinstance(e, ServiceException):
            ex: ServiceException = e
            self._logger.error(__name__, ex.get_detailed_message())
            error = ErrorDTO(ex.error_code, ex.message)
            return jsonify(error.to_dict()), 500
        else:
            tracking_id = uuid.uuid4()
            user_message = f'Tracking Id: {tracking_id}'
            self._logger.error(__name__, user_message, e)
            error = ErrorDTO(None, user_message)
            return jsonify(error.to_dict()), 400

    def before_request(self, *args, **kwargs):
        self._logger.debug(__name__, f'Received GET @{request.url}')
        headers = str(request.headers).replace("\n", "\n\t")
        self._logger.trace(__name__, f'Headers: \n\t{headers}')

    def start(self):
        self._logger.info(__name__, f'Starting API {self._apt_settings.host}:{self._apt_settings.port}')
        self._register_routes()
        # from waitress import serve
        # https://docs.pylonsproject.org/projects/waitress/en/stable/arguments.html
        # serve(self, host=self._apt_settings.host, port=self._apt_settings.port, threads=10, connection_limit=1000, channel_timeout=10)
        wsgi.server(
            eventlet.listen((self._apt_settings.host, self._apt_settings.port)),
            self,
            log_output=False
        )

    def on_connect(self):
        self._logger.info(__name__, f'Client connected')

    def on_disconnect(self):
        self._logger.info(__name__, f'Client disconnected')
